/***********************************************************************

 SV1Codec file I/O routines

 This file is a part of SV1Codec.
 SV1Codec is Copyright (C) 2002 by Marcin Grzegorczyk.

 Permission is granted to copy and redistribute this file under the
 terms of the Open Source MIT License. The full text of the License
 can be found in the file sv1codec.c.

***********************************************************************/

#include "sv1cio.h"
#include "sv1cui.h"
#include "sv1crle.h"

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>




/* Pointer to the stream used by the WriteDataToCurrentFile() callback function */
static struct sStreamInfo *current_write_stream_ptr;




static
int FileIOFail(struct sStreamInfo *const streaminfo)
/*
 Function called when a file I/O operation fails.
 Return 0 on success, nonzero on error.

 If the stream is an ordinary file (as indicated by non-NULL filename
 pointer) attempt to close the stream. If the file is marked as being
 written from scratch, not updated, attempt to delete the file as well.
 Return nonzero if this attempt failed.

 Upon return, if both 'streaminfo->filename' and 'streaminfo->stream'
 are non-NULL, the file close operation failed.
*/
{
  if (streaminfo->filename != NULL) {
    if (fclose(streaminfo->stream) == 0) streaminfo->stream = NULL;  /* don't care much about errors now */

    if (streaminfo->newfileflag) {
      Bool remove_failed;

      UI_IOFail_DeletingFile(streaminfo->filename);
      errno = 0;
      remove_failed = remove(streaminfo->filename) != 0;
      UI_IOFail_DeleteFileResult(remove_failed, errno);
      return remove_failed;
    }
  }
  return 0;
}



PUBLIC_DEFINITION
int WriteData(const void *ptr, size_t n, struct sStreamInfo *const streaminfo)
/*
 Write 'n' bytes from the location pointed to by 'ptr' to a stream.
 Return 0 on success, nonzero on error.

 If an error occurs while writing to an ordinary file, an attempt to
 close and delete the file is taken.
*/
{
  errno = 0;
  if (fwrite(ptr, 1, n, streaminfo->stream) != n) {
    UI_WriteError(streaminfo->filename, errno);
    FileIOFail(streaminfo);
    return EOF;
  }
  return 0;
}



static
int WriteDataToCurrentFile(const void *ptr, size_t n)
/*
 Callback function, needed by the RLE_Decode() and RLE_Encode() functions.

 Write 'n' bytes from the location pointed to by 'ptr' to a stream.
 Return 0 on success, nonzero on error.
*/
{
  assert(current_write_stream_ptr != NULL);
  return WriteData(ptr, n, current_write_stream_ptr);
}



PUBLIC_DEFINITION
char *SearchGameFile(
  const struct sOpParameters *const opp,
  const struct sFilesInfo *const finfo,
  const char *const title)
/*
 Search a path defined in 'finfo->outfname' for a savegame or scenario
 (basing on various parameters in '*opp' and '*finfo') which name
 matches 'title'. If none is found, assign a new filename for the game.

 The path must either be an empty string or a valid directory path,
 including the final path separator character. There must be enough place
 beyond the terminating NUL to append a 9-character file name.

 Return 'finfo->outfname' filled in with the full file & path name,
 or NULL if no match has been found and all possible game filenames are
 already in use.
*/
{
  FILE *file;
  char *tsp = strchr(finfo->outfname, '\0');
  char *const fnn = tsp + 3;  /* A 2-digit savegame file number will be placed there */
  const size_t title_length = opp->sv_format.title_length;
  enum eSVVersion sv_version = opp->sv_version;
  char titlebuf[TTx_MAX_TITLE_LENGTH + 2];
  unsigned number;
  unsigned number_1 = 0;  /* first 'free slot' (unused file name) */
  size_t bytes_read;

  assert(sv_version != SV_VER_UNKNOWN);
  assert(title_length && title_length <= TTx_MAX_TITLE_LENGTH);
  assert(finfo->save_ext[3] == '\0');

  if (sv_version == SV_VER_TTp_DEMO) sv_version = SV_VER_TTO;  /* title-based detection can't distinguish between these two */

  /* append file name to the path */
  *tsp++ = 'T';
  *tsp++ = 'R';
  *tsp++ = finfo->save_letter;
  *tsp++ = '?';
  *tsp++ = '?';
  *tsp++ = '.';
  strcpy(tsp, finfo->save_ext);

  UI_SearchForGameMsg(finfo->outfname);
  for (number = 0; number < 100; number++) {
    fnn[0] = (char)((number / 10u) + '0');
    fnn[1] = (char)((number % 10u) + '0');

    /* try to open the file for reading */
    file = fopen(finfo->outfname, "rb");
    if (file) {
      if (number_1 == number) number_1++;  /* file with this name exists */
      /* try to read the title and title checksum fields */
      bytes_read = fread(titlebuf, 1, title_length + 2, file);
      fclose(file);
      if (bytes_read == title_length + 2) {
        /* check if savegame version and title match */
        if (DetectTTByTitle((unsigned char *)titlebuf, SIZE_MAX) == sv_version) {
          if (strncmp(titlebuf, title, title_length) == 0) {
            UI_GameFoundMsg(finfo->outfname);
            return finfo->outfname;
          }
        }
      }
    }
  }
  UI_GameNotFoundMsg();

  if (number_1 > 99) return NULL;  /* no free slots */

  fnn[0] = (char)((number_1 / 10u) + '0');
  fnn[1] = (char)((number_1 % 10u) + '0');
  return finfo->outfname;
}



static
int AllocAndLoadOpenInFile(struct sFilesInfo *const finfo, FILE *infile)
/*
 Allocate buffer and load the source file after it has been opened

 Return 0 on success, nonzero on error
*/
{
  long infsize;

  /* Try to get the size of the file */
  errno = 0;
  if (fseek(infile, 0, SEEK_END)) {
    UI_CannotGetFileSize(errno);
    return 2;
  }
  errno = 0;
  if ((infsize = ftell(infile)) == -1L) {
    /* oops, something strange has happened */
    UI_CannotGetFileSize(errno);
    return 3;
  }

  UI_FileSizeMsg(infsize);

  finfo->buf.len = (size_t)infsize;
  if(finfo->buf.len != infsize) {
    /*
     Note: we're comparing an unsigned value with a signed one here. This is OK.
     It has to be this way because size_t in principle may be as short as
     unsigned short, or longer than unsigned long long. In addition, ftell()
     is permitted to return negative values on success.

     In any case, if the values compare inequal, it's an unambiguous indication
     that the value returned by ftell() is out of the range of size_t.
    */
    UI_FileTooBig();
    return 4;
  }

  if (infsize < SVx_MIN_FILE_SIZE) {
    UI_FileTooSmall();
    return 5;
  }

  finfo->buf_data_type = BUFDATA_UNKNOWN;

  /* Allocate the buffer */
  finfo->buf.ptr = malloc(finfo->buf.len);
  if (finfo->buf.ptr == NULL) {
    UI_OutOfMemForFileBuf();
    return 6;
  }

  rewind(infile);
  UI_LoadingMsg();

  /* Read the entire file at once */
  errno = 0;
  if (fread(finfo->buf.ptr, 1, finfo->buf.len, infile) != finfo->buf.len) {
    UI_ErrorReadingFile(errno);
    free(finfo->buf.ptr);
    finfo->buf.ptr = NULL;
    return 7;
  }

  UI_LoadedMsg();
  return 0;
}



PUBLIC_DEFINITION
int LoadInFile(struct sFilesInfo *const finfo)
/*
 Open and load the source file

 Return 0 on success, nonzero on error
*/
{
  FILE *infile;
  int ret;

  UI_InputFileMsg(finfo->infname);

  errno = 0;
  infile = fopen(finfo->infname, "rb");
  if (infile == NULL) {
    UI_InFileOpenFailed(errno);
    return 1;
  }

  ret = AllocAndLoadOpenInFile(finfo, infile);

  fclose(infile);
  return ret;
}



PUBLIC_DEFINITION
int CreateOrOpenOutputFile(const Bool create, struct sStreamInfo *const outs)
/*
 Create a new output file if 'create' is TRUE; open an existing file
 for update otherwise. The 'filename' member of the 'outs' structure
 must be initialized to point to the file name to open or create. The
 stdio handle is returned in the 'stream' member.

 Return 0 on success, nonzero on error.
*/
{
  (create ? UI_SavingAsMsg : UI_UpdatingFileMsg)(outs->filename);

  errno = 0;
  outs->stream = fopen(outs->filename, create ? "wb" : "rb+");
  if (outs->stream == NULL) {
    UI_OutFileOpenOrCreateFailed(create, errno);
    return 1;
  }

  outs->newfileflag = create;
  return 0;
}



PUBLIC_DEFINITION
int SeekFileOffset(struct sStreamInfo *const streaminfo, long offset)
{
  errno = 0;
  if (fseek(streaminfo->stream, offset, SEEK_SET) != 0) {
    UI_SeekError(streaminfo->filename, errno);
    FileIOFail(streaminfo);
    return EOF;
  }
  return 0;
}



PUBLIC_DEFINITION
int CloseOutputFile(struct sStreamInfo *const outs, const Bool delay_msg)
/*
 Close the output file.

 If an error occurs and 'delay_msg' is true, the actual error message
 is to be postponed until other warnings or errors are reported.

 Return 0 on success, nonzero on error.
*/
{
  errno = 0;
  if (fclose(outs->stream)) {
    UI_FileCloseFailed(errno, 1, !delay_msg);
    return 1;
  }
  outs->stream = NULL;
  return 0;
}



PUBLIC_DEFINITION
int RLEDecodeToStream(
  const unsigned char *const src,
  const size_t size,
  struct sStreamInfo *const outstream)
{
  current_write_stream_ptr = outstream;
  return RLE_Decode(src, size, WriteDataToCurrentFile);
}



PUBLIC_DEFINITION
int RLEEncodeToStream(
  const unsigned char *const src,
  const size_t size,
  struct sStreamInfo *const outstream,
  uint32_t *const chksump,
  const int code_min,
  const int code_max)
{
  current_write_stream_ptr = outstream;
  return RLE_Encode(src, size, WriteDataToCurrentFile, chksump, code_min, code_max);
}

