/***********************************************************************

 SV1Codec miscellaneous routines

 This file is a part of SV1Codec.
 SV1Codec is Copyright (C) 2002 by Marcin Grzegorczyk.
 See sv1codec.c for details.

 The code in this file is explicitly placed in the public domain.
 Use, abuse and redistribute freely.

***********************************************************************/

#include "sv1cmisc.h"

#include <string.h>
#include <ctype.h>
#include <assert.h>


/********** Global constants **********/

PUBLIC_DEFINITION const char *const g_path_separators = PATH_SEPARATORS;

static const char *const safe_chars = SAFE_CHARS POSIX_SAFE_CHARS;  /* both are string tokens */
static const char *const posix_safe_chars = POSIX_SAFE_CHARS;
static const char HexDigits[] = "0123456789ABCDEF";

/********** End of globals **********/


static
Bool IsSafeChar(char c, const struct sOpParameters *const opp)
/*
 Determine if a nonzero character 'c' is safe to use in file name,
 based on the 'write_escape_mode' and 'use_8bit_filenames' members
 of the structure pointed to by 'opp'.
*/
{
  if (opp->use_8bit_filenames && IS_CHAR_HIGH_SAFE(c)) return 1;
  if ((opp->write_escape_mode & FNAME_ESCAPE_HASH) && c == '#') return 0;
  return strchr((opp->write_escape_mode & FNAME_ESCAPE_POSIX) ? posix_safe_chars : safe_chars, c) != 0;
}



PUBLIC_DEFINITION
void StrFix(
  unsigned char *dest,
  const unsigned char *src,
  const struct sOpParameters *const opp)
/*
 Convert the string pointed to by 'src' to a filesystem-safe name according
 to the rule specified by the 'write_escape_mode' member of the structure
 pointed to by 'opp', placing the result at the location pointed to by
 'dest'. For safe operation, the location pointed to by 'dest' must be able
 to hold at least 'MAX_StrFix_ESCAPE_CHARS*strlen(src)' characters.
*/
{
  while(*src) {
    if (!IsSafeChar(*(char *)src, opp)) {
      /* Character is not safe, must be encoded */
      if (opp->write_escape_mode & FNAME_ESCAPE_POSIX) {
        /* Use POSIX-safe "_0xXX" style (where XX is the hexadecimal representation of the character) */
        *dest++ = '_';
        *dest++ = '0';
        *dest++ = 'x';
      }
      else if (opp->write_escape_mode & FNAME_ESCAPE_URL) {
        /* Use URL-style "%XX" */
        *dest++ = '%';
      }
      else if (opp->write_escape_mode & FNAME_ESCAPE_HASH) {
        /* Use "#xXX" style */
        *dest++ = '#';
        *dest++ = 'x';
      }
      else {
        assert(opp->write_escape_mode & FNAME_ESCAPE_AT);
        /* Use "@XX" style (the default) */
        *dest++ = '@';
      }
      *dest++ = HexDigits[*src >> 4];
      *dest++ = HexDigits[*src++ & 0xF];
    }
    else *dest++ = *src++;  /* Character is safe */
  }
  *dest = '\0';
}



static
size_t StrEscSeqPrefixLength(const char *s, unsigned mode)
/*
 Determine the length of the prefix of a hexadecimal escape sequence
 that starts the string pointed to by 's'. Return 0 if the string does
 not begin with such an escape sequence prefix. 'mode' determines
 allowable escape sequence prefixes.
*/
{
  if ((mode & FNAME_ESCAPE_AT) && *s == '@') return 1;
  if ((mode & FNAME_ESCAPE_URL) && *s == '%') return 1;
  if ((mode & FNAME_ESCAPE_HASH) && *s == '#' && s[1] == 'x') return 2;
  if ((mode & FNAME_ESCAPE_POSIX) && *s == '_' && s[1] == '0' && (s[2] == 'x' || s[2] == 'X')) return 3;
  return 0;
}



PUBLIC_DEFINITION
void StrNUnfix(
  unsigned char *dest,
  const unsigned char *src,
  size_t n,
  const struct sOpParameters *const opp)
/*
 Undo the action done by StrFix(); that is, copy the string pointed to by
 'src' to the location pointed to by 'dest', replacing all escape sequences
 (according to the rules specified by the 'read_escape_modes' member of the
 structure pointed to by 'opp') with appropriate characters. The function
 aborts if 'n' characters have been copied, and thus it is not guaranteed
 that the resulting string will be correctly null-terminated.
*/
{
  while (n--) {
    const char *d1, *d2;
    size_t epl = StrEscSeqPrefixLength((const char *)src, opp->read_escape_modes);
    /*
     Note: lowercase hexdigits are permitted in URL escape sequences.
     See RFC 1738, sec. 2.2, par. 3 (page 3).
    */
    if (epl && src[epl] && (d1 = strchr(HexDigits, toupper(src[epl]))) != NULL && src[epl+1] && (d2 = strchr(HexDigits, toupper(src[epl+1]))) != NULL) {
      src += epl + 2;
      *dest++ = (unsigned char)(((unsigned)(d1 - HexDigits) << 4) | (unsigned)(d2 - HexDigits));
    }
    else if ((*dest++ = *src++) == '\0') break;
  }
}



PUBLIC_DEFINITION
Bool StrnEQ(const char *s1, const char *s2, size_t maxlen)
/*
 Compare two strings in case-insensitive mode, up to 'maxlen' characters.
 Return true if strings are the same, false otherwise.
*/
{
  while (maxlen--) {
    if (tolower((unsigned char)*s1) != tolower((unsigned char)*s2++)) return 0;
    if (*s1++ == '\0') return 1;
  }
  return 1;
}



PUBLIC_DEFINITION
uint16_t GetLEInt16(const void *ptr)
/*
 Return a little-endian 16-bit word taken from the address pointed to by 'ptr'
*/
{
  return (uint16_t)(((unsigned)*((const unsigned char *)ptr + 1) << 8) | (unsigned)*(const unsigned char *)ptr);
}



PUBLIC_DEFINITION
uint32_t GetLEInt32(const void *ptr)
/*
 Return a little-endian 32-bit word taken from the address pointed to by 'ptr'
*/
{
  return (
          (
           (
            (
             (
              (uint32_t)*((const unsigned char *)ptr + 3) << 8
             ) | *((const unsigned char *)ptr + 2)
            ) << 8
           ) | *((const unsigned char *)ptr + 1)
          ) << 8
         ) | *(const unsigned char *)ptr;
}



PUBLIC_DEFINITION
void SetLEInt16(void *ptr, uint16_t w)
/*
 Set a little-endian 16-bit word at the address pointed to by 'ptr'
*/
{
  ((unsigned char *)ptr)[0] = (unsigned char)w;
  ((unsigned char *)ptr)[1] = (unsigned char)(w >> 8);
}



PUBLIC_DEFINITION
void SetLEInt32(void *ptr, uint32_t l)
/*
 Set a little-endian 32-bit word at the address pointed to by 'ptr'
*/
{
  ((unsigned char *)ptr)[0] = (unsigned char)l;
  ((unsigned char *)ptr)[1] = (unsigned char)(l >> 8);
  ((unsigned char *)ptr)[2] = (unsigned char)(l >> 16);
  ((unsigned char *)ptr)[3] = (unsigned char)(l >> 24);
}



PUBLIC_DEFINITION
char *StripFilePath(const char *path)
/*
 Strip the path part from a file name;
 return a pointer to the local file name
*/
{
  const char *s;

  while ((s = strpbrk(path, g_path_separators)) != NULL) path = s + 1;
  return (char *)path;
}

